#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Version 0.2 du 24/05/2019


####
#    NB : Ce script est prévu pour fonctionner sous Linux essentiellement (mais doit être adaptable
#    assez facilement pour d'autres OS...)
#
#    Pré-requis : installation de python3-dnspython (sudo apt-get install python3-dnspython sur
#    debian et dérivées)
#
#    Ce script permet de mettre à jour toutes les entrées présentes dans le fichier hosts du
#    répertoire .../pyHosts
#    Attention, ce fichier est une copie du fichier /etc/hosts réellement utilisé par le système.
#    Une fois la mise à jour effectuée, il faudra remplacer manuellement le fichier /etc/hosts
#    par le fichier .../pyHosts/hosts, à moins qu'un lien symbolique de .../pyHosts/hosts vers
#    /etc/hosts n'ai été créé avec les droits d'écriture qui vont bien...
#
#    Ce script permettra aussi l'ajout de l'historique de navigation du navigateur Brave au
#    fichier hosts. Cet historique est obtenu avec l'extension 'History Export' :
#    --> https://chrome.google.com/webstore/detail/history-export/lpmoaclacdaofhlijejogfldmgkdlglj
#    Cette extension permet de sauver l'historique dans un fichier history.json. Il faut sauver ce
#    fichier dans le répertoire .../pyHosts
#    A la fin du script, ce fichier history.json sera effacé.
#
#    pyHosts.py peut aussi ajouter les entrées des connexions vues par tcpdump. Au préalable, avant
#    de lancer le script pyHosts.py, il faudra démarrer le process tcpdump dans un terminal, à
#    partir du répertoire .../pyHosts 
#                        --> sudo tcpdump -tq tcp > tcpdump.log
#    Cela produira le fichier .../pyHosts/tcpdump.log, qui mémorisera toutes les connexions de
#    votre navigation internet tant que vous n'arrêterez pas ce process. Quand vous déciderez de
#    l'arrêter, vous pourrez alors mettre à jour le fichier hosts en lançant le script
#    .../pyHosts/pyHosts.py. Attention, pour lancer le script pyHosts.py, tcpdump doit
#    impérativement être arrêté ! A la fin du script, le fichier tcpdum.log sera effacé.
#
#    NB : Ce script a été testé sous python3.6.
####




import os
import time
import dns.resolver



def removeRNFromListe(listName):
    ''' Permet de supprimer les \r\n des fins de lignes avant ajout à la stringlist '''
    newList = []
    for oneLine in listName:
        if oneLine.endswith("\r\n"):
            oneLine = oneLine[0:(len(oneLine)-2)]
        if oneLine.endswith("\n"):
            oneLine = oneLine[0:(len(oneLine)-1)]
        if len(oneLine) >0:
            newList.append(oneLine)
    return newList


def readFileLines(filename):
    ''' Ouverture du fichier filename et stockage des lignes lues dans newList'''
    if os.path.exists(filename):
        myFile = open(filename, 'r')
        oneStringList = myFile.readlines()
        myFile.close()
        newList = removeRNFromListe(oneStringList)
        return newList
    else:
        return ("Le fichier "+filename+" n'existe pas!")


def writeFile(filename, lines):
    ''' Ecriture des lignes lines dans le fichier filename '''
    myFile = open(filename, 'w')
    for line in lines:
        oneStr = line + "\r\n"
        myFile.write(oneStr)
    myFile.close()
    return ("Écriture du fichier hosts OK")




###################### Lancement du programme principal
#
if __name__ == "__main__":

    tDepart = time.time()
    currentWorkingDir = os.getcwd()
    currentHostsFilePath = currentWorkingDir + '/hosts'
    currentBraveHistoryFilePath = currentWorkingDir + '/history.json'
    currentTcpdumpFilePath = currentWorkingDir + '/tcpdump.log'

    #### Lecture/chargement du fichier history.json -> Historique de navigation du navigateur Brave
    #
    returnedObject = readFileLines(currentBraveHistoryFilePath)
    braveUrlList = []
    if "<class 'list'>" in str(type(returnedObject)):
        for oneLine in returnedObject:
            if 'https://' in oneLine:
                sep1 = oneLine.split('https://')
                myUrl = sep1[1].split('/')[0]
                if len(myUrl.split('.')) != 4:
                    braveUrlList.append(myUrl)
                else:
                    try:
                        nb1 = int(myUrl.split('.')[0])
                        nb2 = int(myUrl.split('.')[1])
                        nb3 = int(myUrl.split('.')[2])
                        nb4 = int(myUrl.split('.')[3])
                    except:
                        braveUrlList.append(myUrl)
            if 'http://' in oneLine:
                sep1 = oneLine.split('http://')
                myUrl = sep1[1].split('/')[0]
                if len(myUrl.split('.')) != 4:
                    braveUrlList.append(myUrl)
                else:
                    try:
                        nb1 = int(myUrl.split('.')[0])
                        nb2 = int(myUrl.split('.')[1])
                        nb3 = int(myUrl.split('.')[2])
                        nb4 = int(myUrl.split('.')[3])
                    except:
                        braveUrlList.append(myUrl)
        braveUrlSet = set(braveUrlList)
        braveUrlList = list(braveUrlSet)

    #### Lecture/chargement du fichier tcpdump.log - Ce fichier aura eventuellement été
    #    créé par la commande "sudo tcpdump -tq tcp > tcpdump.log" à partir du répertoire
    #    courant .../pyHosts
    #
    returnedObject = readFileLines(currentTcpdumpFilePath)
    print(type(returnedObject))
    tcpdumpUrlList = []
    if "<class 'list'>" in str(type(returnedObject)):
        for oneLine in returnedObject:
            if not 'IP This.' in oneLine:
                sep1 = oneLine.split('.http')
                myUrl = sep1[0][3:]
                if len(myUrl.split('.')) != 4 and not ' > ' in myUrl:
                    tcpdumpUrlList.append(myUrl)
                else:
                    try:
                        nb1 = int(myUrl.split('.')[0])
                        nb2 = int(myUrl.split('.')[1])
                        nb3 = int(myUrl.split('.')[2])
                        nb4 = int(myUrl.split('.')[3])
                    except:
                        if not ' > ' in myUrl:
                            tcpdumpUrlList.append(myUrl)
        tcpdumpUrlSet = set(tcpdumpUrlList)
        tcpdumpUrlList = list(tcpdumpUrlSet)

    #### Chargement du fichier hosts -> Attention, ce fichier hosts est une copie du fichier
    #    /etc/hosts (attention aux droits d'écriture) ! CE FICHIER DOIT TOUJOURS EXISTER !!!
    #
    hostsUrlList = []
    listIPV4_127 = []
    listIPV6 = []
    newLines = []
    myLines = readFileLines(currentHostsFilePath)
    for oneLine in myLines:
        myIP = oneLine.split()[0]
        myHost = oneLine.split()[1]
        if '127.0.0.1' in myIP or '127.0.1.1' in myIP or '###BEGIN' in myIP:
            listIPV4_127.append(oneLine)
        if '###END' in oneLine or '# ' in oneLine:
            listIPV6.append(oneLine)
        if not '127.0.0.' in myIP and not '127.0.1.' in myIP and not '#' in myIP:
            hostsUrlList.append(myHost)

    #### Ajouts des nouvelles entrées en provenance de l'historique de Brave et de tcpdump à la
    #    liste des DNS à mettre à jour dans hostsUrlList - Suppression des URLs déjà présentes 
    #    dans listIPV4_127 de la liste hostsUrlList (si ces URLs sont dans listIPV4_127, c'est 
    #    que l'on souhaite les filtrer, il ne faut pas qu'elle soit dans la liste des adresses
    #    à mettre à jour!)
    #
    nbLinesBeforeUpdate = len(hostsUrlList)
    hostsUrlList.extend(braveUrlList)
    hostsUrlList.extend(tcpdumpUrlList)
    hostsUrlSet = set(hostsUrlList)
    hostsUrlList = list(hostsUrlSet)
    hostsUrlList.sort()
    toRemoveIndex = []
    for oneL127 in listIPV4_127:
        try:
            myHost127 = oneL127.split()[1]
            for theIndex, oneLUrlList in enumerate(hostsUrlList):
                myHostUrlList = oneLUrlList.split()[1]
                if myHost127 in myHostUrlList and myHostUrlList in myHost127:
                    toRemoveIndex.append(theIndex)
        except:
            pass
    toRemoveIndex.sort(reverse=True)
    for iddx in toRemoveIndex:
        del myHostUrlList[iddx]
    nbLines = len(hostsUrlList)
    newEntriesInHosts = nbLines - nbLinesBeforeUpdate
    listIPV4_127.sort()
    listIPV4_127.append(listIPV4_127[0])
    del listIPV4_127[0]
    newLines.extend(listIPV4_127)

    #### Mise à jour des DNS et sauvegarde du fichier hosts
    #
    for index, oneHost in enumerate(hostsUrlList):
        try:
            answ = dns.resolver.query(oneHost, 'A')
            rdata = answ[0]
            newLines.append(str(rdata.address)+' '+oneHost)
            print (index,'/', nbLines, ": ", rdata.address, oneHost)
        except:
            print('Erreur sur résolution DNS de :', oneHost)
    newLines.extend(listIPV6)
    print()
    print('Mise à jour du fichier hosts terminée.', str(newEntriesInHosts),
          'entrées ont été ajoutées au fichier hosts')
    print()
    status = writeFile(currentHostsFilePath, newLines)
    print(status)
    print()
    tUpdate = int(time.time() - tDepart)
    nbMin = int(tUpdate / 60)
    nbSec = int(tUpdate % 60)
    print('Mise à jour effectuée en', str(nbMin), 'min et', str(nbSec), 'sec')
    print()
    print()
    try:
        os.remove(currentBraveHistoryFilePath)
        os.remove(currentTcpdumpFilePath)
    except:
        '''Rien à faire, les fichiers history.json et/ou tcpdump.log auront été
        effacés si besoin'''
        pass
